const std = @import("std");
const zmpl = @import("src/jetzig/zmpl.zig");

pub fn build(b: *std.Build) !void {
    const target = b.standardTargetOptions(.{});
    const optimize = b.standardOptimizeOption(.{});

    const lib = b.addStaticLibrary(.{
        .name = "jetzig",
        .root_source_file = .{ .path = "src/main.zig" },
        .target = target,
        .optimize = optimize,
    });

    b.installArtifact(lib);

    const exe = b.addExecutable(.{
        .name = "jetzig",
        .root_source_file = .{ .path = "src/main.zig" },
        .target = target,
        .optimize = optimize,
    });

    b.installArtifact(exe);

    const jetzig_module = b.createModule(.{ .source_file = .{ .path = "src/jetzig.zig" } });
    exe.addModule("jetzig", jetzig_module);
    lib.addModule("jetzig", jetzig_module);
    try b.modules.put(b.dupe("jetzig"), jetzig_module); // FIXME: Is this needed?

    var templates = std.ArrayList(TemplateDef).init(b.allocator);
    try compileTemplates(b, &templates);
    var dir = std.fs.cwd();
    var file = try dir.createFile("src/app/views/manifest.zig", .{ .truncate = true });
    try file.writeAll("pub const routes = .{\n");
    const routes = try findViews(b.allocator, .zig);
    for (routes.items) |route| {
        try file.writeAll(try std.fmt.allocPrint(b.allocator, "  @import(\"{s}\"),\n", .{route}));
        std.debug.print("[jetzig] Imported view: {s}\n", .{route});
    }

    try file.writeAll("};\n");
    try file.writeAll("pub const templates = struct {\n");
    for (templates.items) |template| {
        const module = b.createModule(.{ .source_file = .{ .path = template.path } });
        exe.addModule(template.name, module);
        lib.addModule(template.name, module);
        try file.writeAll(try std.fmt.allocPrint(
            b.allocator,
            "  pub const {s} = @import(\"{s}\");\n",
            .{ template.name, template.relpath },
        ));
    }
    try file.writeAll("};\n");
    file.close();

    const run_cmd = b.addRunArtifact(exe);

    run_cmd.step.dependOn(b.getInstallStep());
    if (b.args) |args| run_cmd.addArgs(args);

    const run_step = b.step("run", "Run the app");
    run_step.dependOn(&run_cmd.step);

    const main_tests = b.addTest(.{
        .root_source_file = .{ .path = "src/main.zig" },
        .target = target,
        .optimize = optimize,
    });

    const run_main_tests = b.addRunArtifact(main_tests);

    const test_step = b.step("test", "Run library tests");
    test_step.dependOn(&run_main_tests.step);
}

fn compileTemplates(b: *std.Build, array: *std.ArrayList(TemplateDef)) !void {
    const paths = try findViews(b.allocator, .zmpl);
    const base_path = try std.fs.path.join(b.allocator, &[_][]const u8{ "src", "app", "views" });
    const dir = try std.fs.cwd().openDir(base_path, .{});
    for (paths.items) |path| {
        const output_path = try std.fs.path.join(
            b.allocator,
            &[_][]const u8{
                std.fs.path.dirname(path) orelse "",
                try std.mem.concat(b.allocator, u8, &[_][]const u8{
                    ".",
                    std.fs.path.basename(path),
                    ".compiled.zig",
                }),
            },
        );
        var file = try dir.openFile(path, .{});
        const size = (try file.stat()).size;
        const buffer = try b.allocator.alloc(u8, size);
        const content = try dir.readFile(path, buffer);
        var template = zmpl.Template.init(b.allocator, path, content);
        const output = try template.compile();

        const output_file = try dir.createFile(output_path, .{ .truncate = true });
        try output_file.writeAll(output);
        output_file.close();
        try array.append(.{
            .path = try dir.realpathAlloc(b.allocator, output_path),
            .relpath = output_path,
            .name = try template.identifier(),
        });

        std.debug.print("[zmpl] Compiled template: {s}\n", .{path});
    }
}

const TemplateDef = struct {
    path: []const u8,
    relpath: []const u8,
    name: []const u8,
};

fn findViews(allocator: std.mem.Allocator, view_type: enum { zig, zmpl }) !std.ArrayList([]const u8) {
    var array = std.ArrayList([]const u8).init(allocator);
    const dir = try std.fs.cwd().openIterableDir("src/app/views", .{});
    var walker = try dir.walk(allocator);
    defer walker.deinit();
    while (try walker.next()) |entry| {
        if (entry.kind != .file) continue;
        const extension = std.fs.path.extension(entry.path);
        const basename = std.fs.path.basename(entry.path);
        if (std.mem.eql(u8, basename, "manifest.zig")) continue;
        if (std.mem.startsWith(u8, basename, ".")) continue;
        if (view_type == .zig and !std.mem.eql(u8, extension, ".zig")) {
            continue;
        } else if (view_type == .zmpl and !std.mem.eql(u8, extension, ".zmpl")) {
            continue;
        }
        try array.append(try allocator.dupe(u8, entry.path));
    }
    return array;
}
